Attribute VB_Name = "modJPG"
'JpgView for Visual Basic 5.0 and up
'
'Copyright 2002 by Dmitry Brant, All Rights Reserved.

'Version 1.5, Last modified 08-22-2002
'
'me@dmitrybrant.com
'
'http://www.dmitrybrant.com

'Feel free to make improvements or optimizations to the program.
'If you do, please tell me or show me what you have done.

'This program is FREEWARE. Use freely, but give me credit
'where credit is due. (this program is NOT in the
'public domain). If you will use this source code in
'your software, the copyright notice stated above
'must be included in your source code and/or in the
'Help/About box of your program.

'The SOURCE CODE of this program, or portions of it, may not
'be redistributed, either by itself or in a compilation
'package, without express authorization from the author.

'ALL PRODUCT NAMES MENTIONED IN THIS SOFTWARE ARE TRADEMARKS
'OR REGISTERED TRADEMARKS OF THEIR RESPECTIVE OWNERS.

'THIS SOFTWARE IS DISTRIBUTED "AS IS". THE SOFTWARE IS
'DISTRIBUTED WITH NO WARRANTY, EIHER EXPRESSED
'OR IMPLIED, INCLUDING, BY WAY OF EXAMPLE THE IMPLIED
'WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
'PURPOSE. THE AUTHOR WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES,
'LOSS OF PROFITS OR ANY OTHER KIND OF LOSS RESULTING FROM THE
'USE OR MISUSE OF THIS SOFTWARE.

'Using this software signifies acceptance of these terms and
'conditions. If you do not agree with these terms, you must remove
'this software from your storage device(s) and cease to use it.

'Dmitry Brant is not affiliated with the JPEG committee or
'the Independent JPEG Group (IJG). No copyright or
'patent violations intended. This program does not
'support the full JPEG specification:  it does not support
'progressive encoding and arithmetic encoding.
'--------------------------------------------------------
'
'MODULE EXPORTS:
'Sub DoJPG(filename, object)
'    filename = name of JPG image
'    object = the object on which to draw image (i.e. Form1)
'

Option Explicit
DefInt A-Z

Private Type JpegType           'some type definitions (for coherence)
    Rows As Integer             'image height
    Cols As Integer             'image width
    SamplesY As Integer         'sampling ratios
    SamplesCbCr As Integer
    QuantTableY As Integer      'quantization table numbers
    QuantTableCbCr As Integer
    HuffDCTableY As Integer     'huffman table numbers
    HuffDCTableCbCr As Integer
    HuffACTableY As Integer
    HuffACTableCbCr As Integer
    NumComp As Integer          'number of components
End Type

Private Type HuffmanEntry               'a type for huffman tables
    Index As Long
    Code As Integer
    Length As Integer
End Type

Declare Function SetPixel Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal crColor As Long) As Long

'a few global variables
Private curByte As Integer, curBits As Integer, jfile As Integer
Private EOI As Integer, DCTables As Integer, ACTables As Integer, QTables As Integer
Private Image As JpegType
'All large arrays are now dynamic
Private HuffmanDC() As HuffmanEntry
Private HuffmanAC() As HuffmanEntry
Private Dct() As Integer
Private Zig1() As Integer, Zig2() As Integer, ZigIndex As Byte
Private ScanVector() As Byte
Private findex As Long, flen As Long

Private Sub Read(n1 As Byte, n2 As Byte)
    n1 = Zig1(ZigIndex)
    n2 = Zig2(ZigIndex)
    ZigIndex = ZigIndex + 1
End Sub

Private Function Decode(inArray() As HuffmanEntry, hnum As Integer) As Integer
    Dim n1 As Byte, n2 As Byte, i As Integer, l As Integer
    Dim CurVal As Long
    Dim MatchFound As Integer
    If GetByte = 255 Then
        n1 = GetByte
        If n1 >= &HD0 And n1 <= &HD7 Then
            n2 = 2 ^ curBits - 1
            findex = findex - 2
            If curByte And n2 = n2 Then    'if the remaining bits are 1
                EOI = 1
                Decode = 0
                Exit Function
            End If
        Else
            findex = findex - 2
        End If
    Else
        findex = findex - 1
    End If
    CurVal = 0
    MatchFound = -1
    For l = 1 To 16    'cycle through 16 possible Huffman lengths
        CurVal = CurVal * 2 + NextBit
        If EOI Then Exit Function
        For i = 0 To 255              'look for a match in the Huffman table
            If inArray(hnum, i).Length > l Then Exit For
            If inArray(hnum, i).Length = l Then
                If inArray(hnum, i).Index = CurVal Then MatchFound = i: Exit For
            End If
        Next i
        If MatchFound > -1 Then Exit For
    Next l
    If MatchFound = -1 Then
        Beep
        Decode = -1
        Exit Function
    End If
    Decode = inArray(hnum, MatchFound).Code  'return the appropriate code
End Function

Private Sub GetBlock(vector() As Integer, HuffDC() As HuffmanEntry, HuffDCNum As Integer, HuffAC() As HuffmanEntry, HuffACNum As Integer, Quant() As Integer, QuantNum As Integer, dcCoef As Integer)
    Dim d As Integer, XPos As Byte, YPos As Byte, Sum As Integer
    Dim bits As Byte, zeros As Byte, bitVal As Integer, ACCount As Byte
    Dim j As Integer, X As Integer, Y As Integer, v As Integer, u As Integer
    Dim temp As Integer, temp0 As Byte, Add1 As Integer
    
    ReDim array2(0 To 7, 0 To 7) As Integer
    EOI = 0
    temp0 = Decode(HuffDC(), HuffDCNum)   'Get the DC coefficient
    If EOI Then d = 0
    dcCoef = dcCoef + ReceiveBits(temp0)
    array2(0, 0) = dcCoef * Quant(QuantNum, 0, 0)
    XPos = 0: YPos = 0
    ZigIndex = 1
    ACCount = 1
    Do
        d = Decode(HuffAC(), HuffACNum)
        If EOI Then d = 0
        zeros = d \ 16
        bits = d And 15
        bitVal = ReceiveBits(bits)
        If zeros = 0 And bits = 0 Then   'EOB Encountered
            Exit Do
        ElseIf zeros = 15 And bits = 0 Then  'ZRL encountered
            ZigIndex = ZigIndex + 15
            ACCount = ACCount + 16
        Else
            ZigIndex = ZigIndex + zeros
            ACCount = ACCount + zeros
            If ACCount >= 64 Then Exit Do
            
            Read XPos, YPos
            array2(XPos, YPos) = bitVal * Quant(QuantNum, XPos, YPos)
            ACCount = ACCount + 1
        End If
        If ACCount >= 64 Then Exit Do
    Loop
    If HuffDCNum = Image.HuffDCTableY Then Add1 = 128
    For X = 0 To 7            'the IDCT routine (pretty fast)
        For Y = 0 To 7
            Sum = 0
            For v = 0 To 7
                For u = 0 To 7
                    temp = array2(u, v)
                    If temp Then temp = temp * Dct(X, Y, u, v) / 64
                    Sum = Sum + temp
                Next u
            Next v
            vector(X, Y) = Sum + Add1
        Next Y
    Next X
    Erase array2
End Sub

Private Function GetByte() As Integer
    GetByte = ScanVector(findex)
    findex = findex + 1
End Function

Private Function GetHuffTables() As Integer
    Dim HuffAmount(1 To 16) As Integer
    Dim l0 As Long, c0 As Integer, temp0 As Integer, temp1 As Integer, total As Integer
    Dim i As Integer, t0 As Integer, CurNum As Long, CurIndex As Integer, j As Integer
    
    l0 = GetWord
    c0 = 2
    Do
        temp0 = GetByte: c0 = c0 + 1
        t0 = (temp0 And 16) \ 16
        temp0 = temp0 And 15
        Select Case t0
        Case 0        'DC Table
            total = 0
            For i = 1 To 16
                temp1 = GetByte: c0 = c0 + 1
                total = total + temp1
                HuffAmount(i) = temp1
            Next i
            For i = 0 To total - 1
                HuffmanDC(temp0, i).Code = GetByte: c0 = c0 + 1
            Next i
            CurNum = 0
            CurIndex = -1
            For i = 1 To 16
                For j = 1 To HuffAmount(i)
                    CurIndex = CurIndex + 1
                    HuffmanDC(temp0, CurIndex).Index = CurNum
                    HuffmanDC(temp0, CurIndex).Length = i
                    CurNum = CurNum + 1
                Next j
                CurNum = CurNum * 2
            Next i
            DCTables = DCTables + 1
        Case 1
            total = 0
            For i = 1 To 16
                temp1 = GetByte: c0 = c0 + 1
                total = total + temp1
                HuffAmount(i) = temp1
            Next i
            For i = 0 To total - 1
                HuffmanAC(temp0, i).Code = GetByte: c0 = c0 + 1
            Next i
            CurNum = 0
            CurIndex = -1
            For i = 1 To 16
                For j = 1 To HuffAmount(i)
                    CurIndex = CurIndex + 1
                    HuffmanAC(temp0, CurIndex).Index = CurNum
                    HuffmanAC(temp0, CurIndex).Length = i
                    CurNum = CurNum + 1
                Next j
                CurNum = CurNum * 2
            Next i
            ACTables = ACTables + 1
        End Select
    Loop Until c0 >= l0
    GetHuffTables = 1
End Function

Private Function GetImageAttr() As Integer
    Dim temp4 As Long, temp0 As Integer, temp1 As Integer, i As Integer, id As Integer
    
    temp4 = GetWord             'Length of segment
    temp0 = GetByte             'Data precision
    If temp0 <> 8 Then GetImageAttr = 0: Exit Function   'we do not support 12 or 16-bit samples
    Image.Rows = GetWord        'Image Height
    Image.Cols = GetWord        'Image Width
    temp0 = GetByte             'Number of components
    For i = 1 To temp0
        id = GetByte
        Select Case id
            Case 1
                temp1 = GetByte
                Image.SamplesY = (temp1 And 15) * (temp1 \ 16)
                Image.QuantTableY = GetByte
            Case 2, 3
                temp1 = GetByte
                Image.SamplesCbCr = (temp1 And 15) * (temp1 \ 16)
                Image.QuantTableCbCr = GetByte
        End Select
    Next i
    GetImageAttr = 1
End Function

Private Function GetQuantTables(inArray() As Integer) As Integer
    Dim l0 As Long, c0 As Integer, temp0 As Byte
    Dim xp As Byte, yp As Byte, i As Byte
    l0 = GetWord
    c0 = 2
    Do
        temp0 = GetByte: c0 = c0 + 1
        If temp0 And &HF0 Then
            GetQuantTables = 0: Exit Function  'we don't support 16-bit tables
        End If
        temp0 = temp0 And 15
        ZigIndex = 0
        xp = 0: yp = 0
        For i = 0 To 63
            Read xp, yp
            inArray(temp0, xp, yp) = GetByte: c0 = c0 + 1
        Next i
        QTables = QTables + 1
    Loop Until c0 >= l0
    GetQuantTables = 1: Exit Function
End Function

Private Function GetSOI() As Integer
    Dim d As Integer
    findex = 0
    d = 0
    If GetByte = 255 Then
        If GetByte = &HD8 Then d = 1
    End If
    GetSOI = d
End Function

Private Function GetSOS() As Integer
    Dim temp4 As Long, temp0 As Byte, temp1 As Byte, temp2 As Byte
    Dim i As Integer
    
    temp4 = GetWord
    temp0 = GetByte
    If temp0 <> 1 And temp0 <> 3 Then GetSOS = 0: Exit Function
    Image.NumComp = temp0
    For i = 1 To temp0
        temp1 = GetByte
        Select Case temp1
            Case 1
                temp2 = GetByte
                Image.HuffACTableY = temp2 And 15
                Image.HuffDCTableY = temp2 \ 16
            Case 2
                temp2 = GetByte
                Image.HuffACTableCbCr = temp2 And 15
                Image.HuffDCTableCbCr = temp2 \ 16
            Case 3
                temp2 = GetByte
                Image.HuffACTableCbCr = temp2 And 15
                Image.HuffDCTableCbCr = temp2 \ 16
            Case Else
                GetSOS = 0: Exit Function
        End Select 'temp1
    Next i
    findex = findex + 3
    GetSOS = 1
End Function

Private Function GetWord() As Long
    Dim l0 As Long
    l0 = CLng(GetByte) * 256
    l0 = l0 + GetByte
    GetWord = l0
End Function

Private Function NextBit() As Integer
    Dim t0 As Byte, v0 As Byte
    t0 = 2 ^ curBits
    v0 = -((curByte And t0) <> 0)
    curBits = curBits - 1
    If curBits < 0 Then
        curBits = 7: curByte = GetByte
        If curByte = 255 Then
            If GetByte = &HD9 Then EOI = 1: Exit Function
        End If
    End If
    NextBit = v0
End Function

Public Sub DoJPG(FName As String, FObject As Object)
    Dim r As Integer, g As Integer, b As Integer, X As Integer, Y As Integer
    Dim u As Integer, v As Integer, i As Integer, j As Integer, Restart As Long
    Dim t As Single, d As Integer, temp0 As Integer, temp1 As Long
    Dim a As String, theHdc As Long
    Dim XPos As Integer, YPos As Integer, dcY As Integer, dcCb As Integer, dcCr As Integer
    Dim xindex As Integer, yindex As Integer, mcu As Integer
    Dim i2 As Integer, j2 As Integer, cb As Integer, cr As Integer
    Dim xj As Integer, yi As Integer
    
    On Error GoTo DefaultError
    
    ReDim YVector1(0 To 7, 0 To 7) As Integer              '4 vectors for Y attribute
    ReDim YVector2(0 To 7, 0 To 7) As Integer              '(not all may be needed)
    ReDim YVector3(0 To 7, 0 To 7) As Integer
    ReDim YVector4(0 To 7, 0 To 7) As Integer
    ReDim CbVector(0 To 7, 0 To 7) As Integer              '1 vector for Cb attribute
    ReDim CrVector(0 To 7, 0 To 7) As Integer              '1 vector for Cr attribute
    ReDim QuantTable(0 To 1, 0 To 7, 0 To 7) As Integer    '2 quantization tables (Y, CbCr)
    
    ReDim HuffmanDC(0 To 1, 0 To 255) As HuffmanEntry
    ReDim HuffmanAC(0 To 1, 0 To 255) As HuffmanEntry
    ReDim Dct(0 To 7, 0 To 7, 0 To 7, 0 To 7) As Integer
    ReDim Zig1(0 To 63) As Integer
    ReDim Zig2(0 To 63) As Integer
    
    For X = 0 To 7           'Initialize our cosine table (used for DCT)
        For Y = 0 To 7
            For u = 0 To 7
                For v = 0 To 7
                    t = Cos((2 * X + 1) * u * 0.1963495) * Cos((2 * Y + 1) * v * 0.1963495)
                    If u = 0 Then t = t * 0.707107
                    If v = 0 Then t = t * 0.707107
                    'Multiply by 16 to retain precision while staying integer
                    Dct(X, Y, u, v) = t * 16
                Next v
            Next u
        Next Y
    Next X
    
    'store values for Zig-zag reordering
    a = "00011020110203122130403122130405142332415060514233241506071625344352617071625344352617273645546372736455463747566574756657677677"
    For i = 0 To 63
        Zig1(i) = Val(Mid(a, i * 2 + 1, 1))
        Zig2(i) = Val(Mid(a, i * 2 + 2, 1))
    Next i
    a = ""
    
    jfile = FreeFile
    theHdc = FObject.hdc
    
    'The only place where the file is actually read
    Open FName For Binary As #jfile
    If LOF(jfile) = 0 Then Kill FName: GoTo DefaultExit
    flen = LOF(jfile)
    
    ReDim ScanVector(0 To flen + 15) As Byte
    'transfer the entire sucker to a huge array
    Get #jfile, , ScanVector
    Close #jfile           'Done with file
    findex = 0
    
    If GetSOI = 0 Then MsgBox "This is not a valid JPEG/JFIF image.", vbCritical, "Error": GoTo DefaultExit
    
StartOver:
    
    QTables = 0                   'Initialize some checkpoint variables
    ACTables = 0
    DCTables = 0
    Restart = 0
    
    Do                                 'Search for markers
        If GetByte = 255 Then          'Marker Found
            d = GetByte
            Select Case d              'which one is it?
                Case &HC0              'SOF0
                    If GetImageAttr = 0 Then MsgBox "Error getting Start Of Frame 0 Marker.", vbCritical, "Error": GoTo DefaultExit
                Case &HC1              'SOF1
                    If GetImageAttr = 0 Then MsgBox "Error getting Start Of Frame 1 Marker.", vbCritical, "Error": GoTo DefaultExit
                Case &HC9              'SOF9
                    MsgBox "Arithmetic Coding Not Supported.", vbCritical, "Error": GoTo DefaultExit
                Case &HC4              'DHT
                    If ACTables < 2 Or DCTables < 2 Then
                        If GetHuffTables = 0 Then MsgBox "Error getting Huffman tables.", vbCritical, "Error": GoTo DefaultExit
                    End If
                Case &HCC              'DAC
                    MsgBox "Arithmetic Coding Not Supported.", vbCritical, "Error": GoTo DefaultExit
                Case &HDA              'SOS
                    If GetSOS = 0 Then MsgBox "Error getting SOS marker.", vbCritical, "Error": GoTo DefaultExit
                    If (DCTables = 2 And ACTables = 2 And QTables = 2) Or Image.NumComp = 1 Then
                        EOI = 0
                        Exit Do        'Go on to secondary control loop
                    Else
                        MsgBox "Unexpected file format.", vbCritical, "Error": GoTo DefaultExit
                    End If
                Case &HDB              'DQT
                    If QTables < 2 Then
                        If GetQuantTables(QuantTable()) = 0 Then MsgBox "Error getting quantization tables.", vbCritical, "Error": GoTo DefaultExit
                    End If
                Case &HDD              'DRI
                    Restart = GetWord
                Case &HE0              'APP0
                    temp1 = GetWord    'Length of segment
                    findex = findex + 5
                    temp0 = GetByte    'Major revision
                    temp0 = GetByte    'Minor revision
                    temp0 = GetByte    'Density definition
                    temp0 = GetByte    'X-Density
                    temp0 = GetByte    'Y-Density
                    temp0 = GetByte    'Thumbnail Width
                    temp1 = GetByte    'Thumbnail Height
                Case &HFE              'COM
                    findex = findex + GetWord - 2
            End Select 'd
        End If 'Marker found
    Loop Until findex >= flen
    
    XPos = 0: YPos = 0               'Initialize active variables
    dcY = 0: dcCb = 0: dcCr = 0
    xindex = 0: yindex = 0: mcu = 0
    r = 0: g = 0: b = 0
    
    curBits = 7          'Start with the seventh bit
    curByte = GetByte    'Of the first byte
    
    If findex >= flen Then GoTo DefaultExit
    
    'resize the given control to image's dimensions
    FObject.Left = 0
    FObject.Top = 0
    FObject.Width = Image.Cols
    FObject.Height = Image.Rows


    Select Case Image.NumComp        'How many components does the image have?
    Case 3                           '3 components (Y-Cb-Cr)
        Select Case Image.SamplesY   'What's the sampling ratio of Y to CbCr?
        Case 4                       '4 pixels to 1
            Do                       'Process 16x16 blocks of pixels
                GetBlock YVector1(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock YVector2(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock YVector3(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock YVector4(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock CbVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCb
                GetBlock CrVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCr
                'YCbCr vectors have been obtained
                For i = 0 To 7           'Draw top left 8x8 pixels
                    For j = 0 To 7
                        Y = YVector1(i, j)
                        i2 = i \ 2
                        j2 = j \ 2
                        cb = CbVector(i2, j2)
                        cr = CrVector(i2, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                For i = 0 To 7           'Draw top right 8x8 pixels
                    For j = 8 To 15
                        Y = YVector2(i, j - 8)
                        i2 = i \ 2
                        j2 = j \ 2
                        cb = CbVector(i2, j2)
                        cr = CrVector(i2, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                For i = 8 To 15          'Draw bottom left 8x8 pixels
                    For j = 0 To 7
                        Y = YVector3(i - 8, j)
                        i2 = i \ 2
                        j2 = j \ 2
                        cb = CbVector(i2, j2)
                        cr = CrVector(i2, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                For i = 8 To 15          'Draw bottom right 8x8 pixels
                    For j = 8 To 15
                        Y = YVector4(i - 8, j - 8)
                        i2 = i \ 2
                        j2 = j \ 2
                        cb = CbVector(i2, j2)
                        cr = CrVector(i2, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                xindex = xindex + 16
                If xindex >= Image.Cols Then xindex = 0: yindex = yindex + 16: mcu = 1: FObject.Refresh: theHdc = FObject.hdc: DoEvents
                If mcu = 1 And Restart <> 0 Then 'Execute the restart interval
                    curByte = GetByte: curByte = GetByte: curByte = GetByte
                    curBits = 7
                    dcY = 0: dcCb = 0: dcCr = 0: mcu = 0 'Reset the DC value
                End If
            Loop Until findex >= flen Or yindex >= Image.Rows
        Case 2           '2 pixels to 1
            Do
                GetBlock YVector1(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock YVector2(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock CbVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCb
                GetBlock CrVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCr
                'YCbCr vectors have been obtained
                For i = 0 To 7             'Draw left 8x8 pixels
                    For j = 0 To 7
                        Y = YVector1(i, j)
                        j2 = j \ 2
                        cb = CbVector(i, j2)
                        cr = CrVector(i, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                For i = 0 To 7             'Draw right 8x8 pixels
                    For j = 8 To 15
                        Y = YVector2(i, j - 8)
                        j2 = j \ 2
                        cb = CbVector(i, j2)
                        cr = CrVector(i, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                xindex = xindex + 16
                If xindex >= Image.Cols Then xindex = 0: yindex = yindex + 8: mcu = 1: FObject.Refresh: theHdc = FObject.hdc: DoEvents
                If mcu = 1 And Restart <> 0 Then 'execute the restart interval
                    curByte = GetByte: curByte = GetByte: curByte = GetByte
                    curBits = 7
                    dcY = 0: dcCb = 0: dcCr = 0: mcu = 0
                End If
            Loop Until findex >= flen Or yindex >= Image.Rows
        Case 1        '1 pixel to 1
            Do
                GetBlock YVector1(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
                GetBlock CbVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCb
                GetBlock CrVector(), HuffmanDC(), Image.HuffDCTableCbCr, HuffmanAC(), Image.HuffACTableCbCr, QuantTable(), Image.QuantTableCbCr, dcCr
                'YCbCr vectors have been obtained
                For i = 0 To 7            'Draw 8x8 pixels
                    For j = 0 To 7
                        Y = YVector1(i, j)
                        i2 = i \ 2
                        j2 = j \ 2
                        cb = CbVector(i2, j2)
                        cr = CrVector(i2, j2)
                        ToRGB Y, cb, cr, r, g, b
                        xj = xindex + j
                        yi = yindex + i
                        If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(r, g, b)
                    Next j
                Next i
                xindex = xindex + 8
                If xindex >= Image.Cols Then xindex = 0: yindex = yindex + 8: mcu = 1: FObject.Refresh: theHdc = FObject.hdc: DoEvents
                If mcu = 1 And Restart <> 0 Then 'execute the restart interval
                    curByte = GetByte: curByte = GetByte: curByte = GetByte
                    curBits = 7
                    dcY = 0: dcCb = 0: dcCr = 0: mcu = 0
                End If
            Loop Until findex >= flen Or yindex >= Image.Rows
        End Select 'Ratio
    Case 1
        Do
            GetBlock YVector1(), HuffmanDC(), Image.HuffDCTableY, HuffmanAC(), Image.HuffACTableY, QuantTable(), Image.QuantTableY, dcY
            'Y vector has been obtained
            For i = 0 To 7           'Draw 8x8 pixels
                For j = 0 To 7
                    Y = YVector1(i, j)
                    If Y < 0 Then Y = 0
                    If Y > 255 Then Y = 255
                    xj = xindex + j: yi = yindex + i
                    If xj < Image.Cols And yi < Image.Rows Then SetPixel theHdc, xj, yi, RGB(Y, Y, Y)
                Next j
            Next i
            xindex = xindex + 8
            If xindex >= Image.Cols Then xindex = 0: yindex = yindex + 8: mcu = 1: FObject.Refresh: theHdc = FObject.hdc: DoEvents
            If mcu = 1 And Restart <> 0 Then 'execute the restart interval
                curByte = GetByte: curByte = GetByte: curByte = GetByte
                curBits = 7
                dcY = 0: mcu = 0
            End If
        Loop Until findex >= flen Or yindex >= Image.Rows
    End Select 'Components
    
    If findex < flen Then
        'If MsgBox("File appears to contain another scan. Continue processing?", vbYesNo, "Question") = vbYes Then
            GoTo StartOver    'more than one scan in file?
        'End If
    End If
    
DefaultExit:
    Erase ScanVector
    Erase YVector1
    Erase YVector2
    Erase YVector3
    Erase YVector4
    Erase CbVector
    Erase CrVector
    Erase QuantTable
    Erase HuffmanDC
    Erase HuffmanAC
    Erase Dct
    Erase Zig1
    Erase Zig2
    Exit Sub
    
DefaultError:
    MsgBox "There was an error reading the file:" + vbCrLf + Error$, vbCritical, "Error"
    Resume DefaultExit
End Sub

Private Function ReceiveBits(cat As Byte) As Integer
    Dim temp0 As Long, i As Byte
    temp0 = 0
    For i = 1 To cat
        temp0 = temp0 * 2 + NextBit
    Next i
    If temp0 >= 2 ^ (cat - 1) Then
        ReceiveBits = temp0
    Else
        ReceiveBits = -(2 ^ cat - 1) + temp0
    End If
End Function

Private Sub ToRGB(y0 As Integer, cb0 As Integer, cr0 As Integer, r0 As Integer, g0 As Integer, b0 As Integer)
    'Do color space conversion from YCbCr to RGB
    r0 = y0 + cr0 * 7 / 5
    g0 = y0 - cb0 * 7 / 20 - cr0 * 18 / 25
    b0 = y0 + cb0 * 40 / 23
    If r0 > 255 Then r0 = 255
    If r0 < 0 Then r0 = 0
    If g0 > 255 Then g0 = 255
    If g0 < 0 Then g0 = 0
    If b0 > 255 Then b0 = 255
    If b0 < 0 Then b0 = 0
End Sub

